package com.yanzuoguang.cloud.aop;

import com.yanzuoguang.cloud.service.TokenServiceCall;
import com.yanzuoguang.token.TokenHelper;
import com.yanzuoguang.util.exception.ExceptionHelper;
import com.yanzuoguang.util.helper.JsonHelper;
import com.yanzuoguang.util.helper.StringHelper;
import com.yanzuoguang.util.vo.Ref;
import com.yanzuoguang.util.vo.ResponseResult;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.Signature;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.List;

/**
 * LogsAspect（接口请求日志切面）
 *
 * @author: Kang
 * @time: 2018年04月25日 11:43
 */
@Aspect
@Component
public class WebAspect extends BaseRequestAspect {

    private static final String TAG = WebAspect.class.getSimpleName();

    @Autowired
    private TokenServiceCall tokenServiceCall;

    @Value("${yzg.reqUrl:order,check,use}")
    private String reqUrl;

    @Value("${yzg.gateway:^.*gateway.*$}")
    private String gateWay;

    /**
     * exec aop point aspect
     */
    @Pointcut("execution(* *..web..*Controller.*(..))")
    public void webAspect() {
    }

    /**
     * 执行环形切面
     *
     * @param joinPoint
     * @return
     */
    @Around(value = "webAspect()")
    public Object requestWebAround(ProceedingJoinPoint joinPoint) throws Throwable {
        boolean clear = requestLogInit();
        // 用户数据库记录
        long start = requestLog(TAG, joinPoint);
        Object result = null;
        Exception ex = null;
        Ref<Boolean> flag = new Ref<>(false);
        boolean isGateWay = isGateWay();
        try {
            if (!isGateWay) {
                tokenServiceCall.tokenInit();
            }
            result = executeMethod(joinPoint);
            if (!isGateWay) {
                tokenFinish(flag);
            }
            return result;
        } catch (Exception e) {
            ex = e;
            result = ExceptionHelper.getError(e);
            if (getReturnType(joinPoint).getTypeName().indexOf(ResponseResult.class.getName()) > -1) {
                return result;
            } else {
                throw e;
            }
        } finally {
            if (!isGateWay) {
                tokenFinish(flag);
            }
            TokenHelper.remove();
            responseLog(clear, TAG, HttpAspectUtil.getHttpRequestUrl(), joinPoint, start, result, ex);
        }
    }

    /**
     * 是否属于网关服务,网关服务不进行监控
     *
     * @return
     */
    private boolean isGateWay() {
        return this.applicationName.matches(this.gateWay);
    }

    /**
     * 执行结束函数
     *
     * @param flag 结束函数
     * @return 结束函数
     */
    private void tokenFinish(Ref<Boolean> flag) {
        if (flag.value == true) {
            return;
        }
        flag.value = true;
        tokenServiceCall.tokenFinish();
    }

    /**
     * 获取返回的至类型
     *
     * @param joinPoint
     * @return
     * @throws NoSuchMethodException
     */
    private Type getReturnType(ProceedingJoinPoint joinPoint) {
        Method m = getMethod(joinPoint);
        Type t = m.getAnnotatedReturnType().getType();
        return t;
    }

    /**
     * 获取返回的至类型
     *
     * @param joinPoint
     * @return
     * @throws NoSuchMethodException
     */
    private Method getMethod(ProceedingJoinPoint joinPoint) {
        //获取返回值类型
        Signature s = joinPoint.getSignature();
        MethodSignature ms = (MethodSignature) s;
        Method m = ms.getMethod();
        return m;
    }

    /**
     * 执行方法
     *
     * @param joinPoint 需要执行的方法
     * @return 返回结果
     * @throws Throwable
     */
    private Object executeMethod(ProceedingJoinPoint joinPoint) throws Throwable {
        List<Object> dataParas = this.getDataParameters(joinPoint.getArgs());
        boolean dataArgs = dataParas.isEmpty();
        Method method = getMethod(joinPoint);
        boolean isUrl = false;
        if (!StringHelper.isEmpty(reqUrl)) {
            String[] urls = reqUrl.split(",");
            for (String url : urls) {
                if (method.getName().matches(url)) {
                    isUrl = true;
                    break;
                }
            }
        }
        if (dataArgs || !isUrl) {
            return joinPoint.proceed();
        } else {
            // 获取请求编号
            Object firstArgs = dataParas.size() > 0 ? dataParas.get(0) : null;
            String reqId = StringHelper.md5(JsonHelper.serialize(firstArgs));
            String reqFull = StringHelper.getId(reqId, HttpAspectUtil.getHttpRequestUrl());
            RequestCacheResult req = cacheResult.get(reqFull, new RequestCacheResult(reqFull));
            // 缓存的键值对
            if (req.getResult() == null) {
                synchronized (req) {
                    if (req.getResult() == null) {
                        try {
                            req.setResult(joinPoint.proceed());
                        } catch (Exception ex) {
                            req.setResult(ex);
                        }
                    }
                }
            }
            Object result = req.getResult();
            if (result instanceof Throwable) {
                throw (Throwable) result;
            }
            return result;
        }
    }
}
